enum CHAT_COLORS
{
	CHAT_Grey,
	CHAT_Green,
	CHAT_Blue,
	CHAT_Red,
	CHAT_Yellow,
	CHAT_Purple,
	CHAT_Orange,
	CHAT_Pink
};

class	_font;														// Text-drawing interface

class	_chat;														// Chat list rendering device

struct	_surface;													// DIRECTDRAWSURFACE wrapper

class	_tileset;													// TileSet

LPDIRECTDRAWSURFACE7 CreateSurface(IDirectDraw7 *,
								  unsigned long, unsigned long);	// Make an offscreen surface of the given dimensions

LPDIRECTDRAWSURFACE7 LoadBitmap(IDirectDraw7 *,
							   const char *,
							   unsigned long &, unsigned long &);	// Make an offscreen surface from a bitmap on disk


//Basic surface

struct _surface
{
	_surface(DXWindow *, const char *);			// Create a surface from picture on disk
	_surface(DXWindow *,
			 unsigned long, unsigned long);		// Create a blank surface of specified dimensions
	~_surface();								// Destroy the surface BTS

	LPDIRECTDRAWSURFACE7 image;					// DirectX Surface handle

	unsigned long width;						// Width in pixels
	unsigned long height;						// Height in pixels
};

LPDIRECTDRAWSURFACE7 CreateSurface(DXWindow * Window, unsigned long Width, unsigned long Height)
{
	DDSURFACEDESC2			ddsd;
	LPDIRECTDRAWSURFACE7	dds;

	/* Create a set of parameters for surface creation */
    ZeroMemory(&ddsd, sizeof(ddsd));
    ddsd.dwSize			= sizeof(ddsd);
    ddsd.dwFlags		= DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth		= Width;
	ddsd.dwHeight		= Height;
    ddsd.ddsCaps.dwCaps	= DDSCAPS_OFFSCREENPLAIN;

	/* Execute creation */
    Window->pDD->CreateSurface(&ddsd, (IDirectDrawSurface7 **) &dds, NULL);

	/* Make a color key range */
	_DDCOLORKEY ddck;
	ddck.dwColorSpaceHighValue	= 0;
	ddck.dwColorSpaceLowValue	= 0;

	/* Set surface color key */
	dds->SetColorKey(DDCKEY_SRCBLT, &ddck);

	return (dds);
}

LPDIRECTDRAWSURFACE7 LoadBitmap(DXWindow * Window, const char * file, unsigned long &width, unsigned long &height)
{
	HBITMAP hbm		= (HBITMAP)LoadImage(NULL, file, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);		// Load DIB from disk into BITMAP object

	if (hbm == NULL)
	{
		MessageBox(NULL, "Unable to load a texture.  I will now crash for you.", file, 0);
	}

	HDC hdcImage	= CreateCompatibleDC(NULL);																			// Create "Device Context" for the DIB
	HBITMAP hbmOld	= (HBITMAP)SelectObject(hdcImage, hbm);																// Bind DC to DIB

		BITMAP bmh;																										// Declare temporary DIB header structure 'bmh'
		GetObject(hbm, sizeof(bmh), &bmh);																				// Overwrite 'bmh' with the DIB header

		LPDIRECTDRAWSURFACE7 dds = CreateSurface(Window, (unsigned long) bmh.bmWidth, (unsigned long) bmh.bmHeight);		// Use dimensions of DIB to initialize surface

		HDC hdc;																										// Declare temporary DC for BitBlt()
		dds->GetDC(&hdc);																								// Requisition DC from surface

			BitBlt(hdc, 0, 0, bmh.bmWidth, bmh.bmHeight, hdcImage, 0, 0, SRCCOPY);										// Blit DIB to surface

		dds->ReleaseDC(hdc);																							// Release surface DC

	SelectObject(hdcImage, hbmOld);																						// Help Windows clean up (Might be unnecessary)
	DeleteDC(hdcImage);																									// Release DIB DC

	width	= (unsigned long) bmh.bmWidth;
	height	= (unsigned long) bmh.bmHeight;

	return (dds);
}

_surface::_surface(DXWindow * Window, const char * file)
{
	image = LoadBitmap(Window, file, width, height);
}

_surface::_surface(DXWindow * Window, unsigned long Width, unsigned long Height)
{
	image = CreateSurface(Window, (width = Width), (height = Height));
}

_surface::~_surface()
{
	if (image)
		image->Release();
}


//Basic tileset

class _tileset
{
public:
	_surface * tiles;							// Handle to the buffer containing tiles

	unsigned long width_in_tiles;				// Number of tiles in the x-direction

	_tileset(DXWindow   *,
			 const char *, unsigned long,
			               unsigned long);		// Load tiles from disk.  Parse with dimensions given
	~_tileset();								// Destroy the surface BTS

	unsigned long tile_width;					// Width of each tile in pixels
	unsigned long tile_height;					// Height of each tile in pixels

	void OpaqueBlit(LPDIRECTDRAWSURFACE7,		// Dx Surface to draw to
			    unsigned long,					// DestinationX
			    unsigned long,					// DestinationY
			    unsigned short);				// Copy specified tile to given surface / Simple full-tile opaque blit

	void TransBlit(LPDIRECTDRAWSURFACE7,		// Dx Surface to draw to
			    unsigned long,					// DestinationX
			    unsigned long,					// DestinationY
			    unsigned short);				// Copy specified tile to given surface / Simple full-tile transparent blit

	void OpaqueBlitEx(LPDIRECTDRAWSURFACE7,		// Surface to draw on
			  unsigned long,					// DestinationX
			  unsigned long,					// DestinationY
			  unsigned long,					// SourceX
			  unsigned long,					// SourceY
			  unsigned long,					// Width
			  unsigned long,					// Height
			  unsigned short);					// Copy specified tile to given surface / Advanced partial-tile blit

	void TransBlitEx(LPDIRECTDRAWSURFACE7,		// Surface to draw on
			  unsigned long,					// DestinationX
			  unsigned long,					// DestinationY
			  unsigned long,					// SourceX
			  unsigned long,					// SourceY
			  unsigned long,					// Width
			  unsigned long,					// Height
			  unsigned short);					// Copy specified tile to given surface / Advanced partial-tile blit
};

_tileset::_tileset(DXWindow * Screen, const char * file, unsigned long TileWidth, unsigned long TileHeight)
{
#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "In _tileset() constructor", "Loading...", 0);
#endif

	tiles = new _surface(Screen, file);

	width_in_tiles = tiles->width / (tile_width = TileWidth);
	tile_height = TileHeight;

#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "Out of _tileset() constructor", "Loading...", 0);
#endif
}

_tileset::~_tileset()
{
	delete tiles;
}

inline void _tileset::OpaqueBlit(LPDIRECTDRAWSURFACE7 dest, unsigned long ScreenX, unsigned long ScreenY, unsigned short tile)
{
	unsigned long x = (tile % width_in_tiles) * tile_width;						// Calculate X offset in tileset from 'tile' number
	unsigned long y = (tile / width_in_tiles) * tile_height;					// Calculate Y offset in tileset from 'tile' number

	RECT r = {x, y, (x + tile_width), (y + tile_height)};						// Generate a BltFast() readable version

	dest->BltFast(ScreenX, ScreenY, tiles->image, &r, DDBLTFAST_NOCOLORKEY);
}

inline void _tileset::TransBlit(LPDIRECTDRAWSURFACE7 dest, unsigned long ScreenX, unsigned long ScreenY, unsigned short tile)
{
	unsigned long x = (tile % width_in_tiles) * tile_width;						// Calculate X offset in tileset from 'tile' number
	unsigned long y = (tile / width_in_tiles) * tile_height;					// Calculate Y offset in tileset from 'tile' number

	RECT r = {x, y, (x + tile_width), (y + tile_height)};						// Generate a BltFast() readable version

	dest->BltFast(ScreenX, ScreenY, tiles->image, &r, DDBLTFAST_SRCCOLORKEY);
}

inline void _tileset::OpaqueBlitEx(LPDIRECTDRAWSURFACE7 dest, unsigned long ScreenX, unsigned long ScreenY, unsigned long SrcX, unsigned long SrcY, unsigned long Width, unsigned long Height, unsigned short tile)
{
	unsigned long x = (tile % width_in_tiles) * tile_width + SrcX;				// Calculate X offset in tileset from 'tile' number
	unsigned long y = (tile / width_in_tiles) * tile_height + SrcY;				// Calculate Y offset in tileset from 'tile' number

	RECT r = {x, y, (x + Width), (y + Height)};									// Generate a BltFast() readable version

	dest->BltFast(ScreenX, ScreenY, tiles->image, &r, DDBLTFAST_NOCOLORKEY);
}

inline void _tileset::TransBlitEx(LPDIRECTDRAWSURFACE7 dest, unsigned long ScreenX, unsigned long ScreenY, unsigned long SrcX, unsigned long SrcY, unsigned long Width, unsigned long Height, unsigned short tile)
{
	unsigned long x = (tile % width_in_tiles) * tile_width + SrcX;				// Calculate X offset in tileset from 'tile' number
	unsigned long y = (tile / width_in_tiles) * tile_height + SrcY;				// Calculate Y offset in tileset from 'tile' number

	RECT r = {x, y, (x + Width), (y + Height)};									// Generate a BltFast() readable version

	dest->BltFast(ScreenX, ScreenY, tiles->image, &r, DDBLTFAST_SRCCOLORKEY);
}


//Basic emulated font

class _font
{
	_tileset * tileset;

public:
	_font(DXWindow *, const char *);
	~_font();

	void DrawOpaqueText(LPDIRECTDRAWSURFACE7 surface, unsigned long x, unsigned long y, const char * msg, BYTE Color);
	void DrawTransText(LPDIRECTDRAWSURFACE7 surface, unsigned long x, unsigned long y, const char * msg, BYTE Color);
};

_font::_font(DXWindow * Window, const char * filename)
{
	tileset	= new _tileset(Window, filename, 8, 8);
}

_font::~_font()
{
	if (tileset)	delete tileset;
}

void _font::DrawOpaqueText(LPDIRECTDRAWSURFACE7 surface, unsigned long x, unsigned long y, const char * msg, BYTE Color)
{
	int i = 0;

	while (msg[i])
	{
		unsigned char c = msg[i];

		if (c >= 32 && c <= 126)
			c -= 32;
		else
			c = 0;

		tileset->OpaqueBlit(surface, x + (i * tileset->tile_width), y, c + (Color * 96));

		i++;
	}
}

void _font::DrawTransText(LPDIRECTDRAWSURFACE7 surface, unsigned long x, unsigned long y, const char * msg, BYTE Color)
{
	int i = 0;

	while (msg[i])
	{
		unsigned char c = msg[i];

		if (c >= 32 && c <= 126)
			c -= 32;
		else
			c = 0;

		tileset->TransBlit(surface, x + (i * tileset->tile_width), y, c + (Color * 96));

		i++;
	}
}


//Basic chat system

struct ChatLine
{
	unsigned char * Message;
	unsigned long	MessageWidth;
	BYTE			Color;
};

class _chat
{
	_font		* font;
	_surface	* buffer;
	DXWindow	* screen;

	unsigned long ScreenX;
	unsigned long ScreenY;
	unsigned long Width;
	unsigned long Height;

	ChatLine MessageList[50];

	unsigned long CurrentMessage;
	unsigned long TotalMessages;
	unsigned long MaxMessages;

public:
	_chat(DXWindow *, unsigned long, unsigned long, unsigned long, unsigned long);
	~_chat();

	void AddMessage(const char * msg, BYTE Color);
	void BlitBuffer();
	void SetFont(_font *);
};

_chat::_chat(DXWindow * nScreen, unsigned long x, unsigned long y, unsigned long w, unsigned long h)
{
#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "In _chat() constructor", "Loading...", 0);
#endif

	screen			= nScreen;

	ScreenX			= x;
	ScreenY			= y;
	Width			= w;
	Height			= h;

	TotalMessages	= 0;
	MaxMessages		= Height >> 3;
	CurrentMessage	= 0xFFFFFFFF;

	buffer = new _surface(screen, Width, Height);

	for (WORD i = 0; i < MaxMessages; i++)
		MessageList[i].Message = NULL;

#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "Out of _chat() constructor", "Loading...", 0);
#endif
}

_chat::~_chat()
{
	delete buffer;

//	for (WORD i = 0; i < MaxMessages; i++)
//		if (MessageList[i].Message)
//			delete MessageList[i].Message;
}

void _chat::AddMessage(const char * msg, BYTE Color)
{
	if (strlen(msg) > 80)
	{
		for (DWORD Offset = 0; Offset < strlen(msg); Offset += 80)
		{
			char * newmsg = new char[81];
			
			if (Offset + 80 < strlen(msg))
			{
				// body
				memcpy(newmsg, &msg[Offset], 80);
				newmsg[80] = 0;
				AddMessage(newmsg, Color);
			}
			else
			{
				// tail
				memcpy(newmsg, &msg[Offset], strlen(msg) - Offset);
				newmsg[strlen(msg) - Offset] = 0;
				AddMessage(newmsg, Color);
			}

			delete newmsg;
		}

		return;
	}

	if (TotalMessages < MaxMessages)
		TotalMessages++;

	CurrentMessage++;
	if (CurrentMessage >= TotalMessages)
		CurrentMessage = 0;

	MessageList[CurrentMessage].Message = new BYTE[strlen(msg) + 1];
	strcpy((char *)MessageList[CurrentMessage].Message, msg);

	font->DrawOpaqueText(buffer->image, 0, CurrentMessage << 3, msg, Color);
	MessageList[CurrentMessage].MessageWidth = strlen(msg) << 3;
	MessageList[CurrentMessage].Color = Color;
}

void _chat::SetFont(_font * nFont)
{
	font = nFont;
}

void _chat::BlitBuffer()
{
	int Parse = CurrentMessage + 1;

	for (WORD i = 1; i <= TotalMessages; i++)
	{
		Parse--;
		if (Parse < 0)
			Parse = TotalMessages - 1;

		RECT r = {0, Parse << 3, MessageList[Parse].MessageWidth, (Parse + 1) << 3};
		screen->Backbuffer->BltFast(ScreenX, ScreenY + (Height) - (i << 3), buffer->image, &r, DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT);
	}
}


// Basic number set

class _numbers
{
	_tileset	* font;
	DXWindow	* screen;

	WORD x, y;
	bool RightJustify;

public:
	_numbers(DXWindow * nScreen, bool nRightJustify)
	{
		screen = nScreen;
		RightJustify = nRightJustify;
	}

	void SetFont(_tileset * NewFont)
	{
		font = NewFont;
	}

	void Move(WORD nx, WORD ny)
	{
		x = nx;
		y = ny;
	}

	void Render(WORD Number)
	{
		WORD Offset = x;

		do
		{
			WORD ThisOne = Number % 10;
			Number = Number / 10;

			if (RightJustify)
				Offset -= (WORD)font->tile_width;
			else
				Offset += (WORD)font->tile_width;

			font->TransBlit(screen->Backbuffer, Offset, y, ThisOne);
		} while (Number);
	}
};


// Basic chat entry box

class ChatEntryBox
{
	char Entered[128];
	long TotalKeys;
	DWORD LastTime;
	DWORD RepeatRate;
	char LastKey;

	_font		* font;
	DXWindow	* screen;

public:
	ChatEntryBox(DXWindow * nScreen, DWORD nRepeatRate)
	{
		screen = nScreen;
		RepeatRate = nRepeatRate;
		TotalKeys = 0;
	}

	void SetFont(_font * nFont)
	{
		font = nFont;
	}

	bool KeyDown(char KeyCode)
	{
		// Delete
		if (KeyCode == 8)
		{
			if (TotalKeys)
			{
				Entered[--TotalKeys] = 0;
			}

			return false;
		}

		// Enter
		if (KeyCode == 13)
		{
			TotalKeys = 0;

			return true;
		}

		// Limit length
		if (TotalKeys >= (screen->Width >> 3))
			return false;

		// Limit repeat rate
		if (KeyCode == LastKey)
		{
			if (GetTickCount() - LastTime < RepeatRate)
				return false;
		}
		else
		{
			LastKey = KeyCode;
		}

		Entered[TotalKeys++] = KeyCode;
		Entered[TotalKeys] = 0;

		LastTime = GetTickCount();

		return false;
	}

	void Render()
	{
		font->DrawTransText(screen->Backbuffer, 0, screen->Height - 8, Entered, CHAT_Grey);
	}
};


// Basic SubSpace-style GUI frame

class _frame
{
	DXWindow * Screen;
	LPDIRECTDRAWSURFACE7 buffer;
	LPDIRECTDRAWSURFACE7 dest;
	DDSURFACEDESC2 ddsd;
	RECT SourceRect;

	DWORD x, y, w, h;

public:
	_frame(DXWindow * nScreen)
	{
		Screen	= nScreen;

		/* Create the backbuffer surface */
	    ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize		= sizeof(ddsd);

		ddsd.dwFlags		= DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT; // | DDSD_PIXELFORMAT
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwWidth		= 1700;
		ddsd.dwHeight		= 1300;

		HRESULT retval = Screen->pDD->CreateSurface(&ddsd, &buffer, NULL);

		if (retval != DD_OK)
		{
			_string S;
			S = "Screen->pDD->CreateSurface() Error # ";
			S += retval;
			S += ": ";
			S += DDErrorString(retval);
			MessageBox(NULL, S.Text, "_radar::_radar()", 0);
		}

		SourceRect.left = 0;
		SourceRect.top = 0;
	}

	~_frame()
	{
		buffer->Release();
	}

	void Move(DWORD nx, DWORD ny)
	{
		x = nx;
		y = ny;
	}

	void Size(DWORD nw, DWORD nh)
	{
		w = nw;
		h = nh;

		SourceRect.right = w;
		SourceRect.bottom = h;

		if (buffer->Lock(NULL, &ddsd, DDLOCK_NOSYSLOCK | DDLOCK_WRITEONLY | DDLOCK_WAIT, NULL) == DD_OK)
		{
			BYTE * mem = (BYTE*)ddsd.lpSurface;

			for (DWORD ix = 0; ix < w; ix++)
			{
				mem[ix + 0 * ddsd.lPitch] = 255;
				mem[ix + h * ddsd.lPitch] = 255;
			}

			for (DWORD iy = 1; iy < (h - 1); iy++)
			{
				mem[0 + iy * ddsd.lPitch] = 255;
				mem[w + iy * ddsd.lPitch] = 255;
			}

			buffer->Unlock(NULL);
		}
	}

	void SetSurface(LPDIRECTDRAWSURFACE7 nDest)
	{
		dest = nDest;
	}

	void Render()
	{
		HRESULT retval = dest->BltFast(x, y, buffer, &SourceRect, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT);

		if (retval != DD_OK)
		{
			_string S;
			S = "dest->BltFast() Error # ";
			S += retval;
			S += ": ";
			S += DDErrorString(retval);
			MessageBox(NULL, S.Text, "_frame::BlitBuffer()", 0);
		}
	}
};


// Radar

class _radar
{
	BYTE * Map;
	DXWindow * Screen;
	LPDIRECTDRAWSURFACE7 buffer;
	DDSURFACEDESC2 ddsd;
	RECT SourceRect;
	_frame * RadarFrame;

	_surface * RightRadarBrace;
	_surface * BottomRadarBrace;

	DWORD LastX, LastY;

public:
	_radar(DXWindow * nScreen)
	{
		Screen	= nScreen;

		/* Create the backbuffer surface */
	    ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize		= sizeof(ddsd);

		ddsd.dwFlags		= DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;	// | DDSD_PIXELFORMAT
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwWidth		= 173;
		ddsd.dwHeight		= 173;

		HRESULT retval = Screen->pDD->CreateSurface(&ddsd, &buffer, NULL);

		if (retval != DD_OK)
		{
			_string S;
			S = "Screen->pDD->CreateSurface() Error # ";
			S += retval;
			S += ": ";
			S += DDErrorString(retval);
			MessageBox(NULL, S.Text, "_radar::_radar()", 0);
		}

		// Pre-render frame
		RadarFrame = new _frame(Screen);
		RadarFrame->SetSurface(buffer);
		RadarFrame->Move(0, 0);
		RadarFrame->Size(173, 173);
		RadarFrame->Render();

		// Pre-render braces
		RightRadarBrace = new _surface(Screen,	"gfx/vframe.bmp");
		BottomRadarBrace = new _surface(Screen,	"gfx/hframe.bmp");

		SourceRect.top = 0;
		SourceRect.left = 0;

		SourceRect.right = 7;
		SourceRect.bottom = 62;
		buffer->BltFast(166, 104, RightRadarBrace->image, &SourceRect, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT);

		SourceRect.right = 75;
		SourceRect.bottom = 7;
		buffer->BltFast(98, 166, BottomRadarBrace->image, &SourceRect, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT);

		SourceRect.right = 173;
		SourceRect.bottom = 173;
	}

	~_radar()
	{
		buffer->Release();
		delete RightRadarBrace;
		delete BottomRadarBrace;
		delete RadarFrame;
	}

	void SetMap(BYTE * nMap)
	{
		Map = nMap;
	}

	void Render()
	{
		DWORD TileX		= Screen->CenterX >> 4;
		DWORD TileY		= Screen->CenterY >> 4;

		if ((TileX != LastX) || (TileY != LastY))
		{
			DWORD StartX	= TRIM(TileX, 82);
			DWORD StartY	= TRIM(TileY, 82);

			if (buffer->Lock(NULL, &ddsd, DDLOCK_NOSYSLOCK | DDLOCK_WRITEONLY | DDLOCK_WAIT, NULL) == DD_OK)
			{
				BYTE * mem = (BYTE*)ddsd.lpSurface;

				for (DWORD x = 0; x < 163; x++)
				for (DWORD y = 0; y < 163; y++)
				{
					DWORD MapX = StartX + x;
					DWORD MapY = StartY + y;

					if ((MapX < 1024) && (MapY < 1024))
					{
						BYTE v = Map[(MapY << 10) + MapX];

						mem[(x + 3) + (y + 3) * ddsd.lPitch] = v;
					}
				}

				buffer->Unlock(NULL);
			}

			LastX = TileX;
			LastY = TileY;
		}

		HRESULT retval = Screen->Backbuffer->BltFast(Screen->Width - 173, Screen->Height - 173, buffer, &SourceRect, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT);

		if ((retval != DD_OK) && (retval != DDERR_INVALIDRECT))
		{
			_string S;
			S = "Screen->Backbuffer->BltFast() Error # ";
			S += retval;
			S += "[";
			S += (int)SourceRect.right;
			S += ", ";
			S += (int)SourceRect.bottom;
			S += "]";
			S += ": ";
			S += DDErrorString(retval);
			MessageBox(NULL, S.Text, "_radar::BlitBuffer()", 0);
		}
	}
};
